(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Abstract_Rewriting2 
imports Abstract_Rewriting
begin

(* this theory is used to collect differences between the official AFP and 
   our version *)


lemma NF_Id_on_fst_image[simp]: "NF (Id_on (fst ` A)) = NF A" by force

lemma fst_image_NF_Id_on[simp]: "fst ` R = Q \<Longrightarrow> NF (Id_on Q) = NF R" by force

lemma NF_empty[simp]: "NF {} = UNIV" by auto

lemma trancl_mono_set: "r \<subseteq> s \<Longrightarrow> r^+ \<subseteq> s^+"
   using trancl_mono by auto

lemma SN_imp_irreflexive: assumes "SN r" shows "(l,l) \<notin> r"
proof
  assume in_gr: "(l,l) \<in> r"
  with `SN r` show False unfolding SN_defs by auto
qed


lemma SN_nat_bounded: "SN {(x,y :: nat). x < y \<and> y \<le> b}" (is "SN ?R")
proof 
  fix f
  assume "\<forall> i. (f i, f (Suc i)) \<in> ?R"
  hence steps: "\<And> i. (f i, f (Suc i)) \<in> ?R" ..
  {
    fix i
    have inc: "f 0 + i \<le> f i"
    proof (induct i)
      case 0 thus ?case by auto
    next
      case (Suc i)
      have "f 0 + Suc i \<le> f i + Suc 0" using Suc by simp
      also have "... \<le> f (Suc i)" using steps[of i]
        by auto
      finally show ?case by simp
    qed
  }
  from this[of "Suc b"] steps[of b]
  show False by simp
qed

lemma first_step: assumes C: "C = A \<union> B" and steps: "(x,y) \<in> C^*"
  and Bstep: "(y,z) \<in> B"
  shows "\<exists> y. (x,y) \<in> A^* O B"
  using steps
proof (induct rule: converse_rtrancl_induct)
  case base
  show ?case using Bstep by auto
next 
  case (step u x)
  from step(1)[unfolded C] 
  show ?case
  proof
    assume "(u,x) \<in> B"
    thus ?thesis by auto
  next
    assume ux: "(u,x) \<in> A"
    from step(3) obtain y where "(x,y) \<in> A^* O B" by auto
    then obtain z where "(x,z) \<in> A^*" and step: "(z,y) \<in> B" by auto
    with ux have "(u,z) \<in> A^*" by auto
    with step have "(u,y) \<in> A^* O B" by auto
    thus ?thesis by auto
  qed
qed

lemma first_step_O: assumes C: "C = A \<union> B" and steps: "(x,y) \<in> C^* O B"
  shows "\<exists> y. (x,y) \<in> A^* O B"
proof -
  from steps obtain z where "(x,z) \<in> C^*" and "(z,y) \<in> B" by auto
  from first_step[OF C this] show ?thesis .
qed

lemma firstStep: assumes LSR: "L = S \<union> R" and xyL: "(x,y) \<in> L^*"
  shows "(x,y) \<in> R^* \<or> (x,y) \<in> R^* O S O L^*"
proof (cases "(x,y) \<in> R^*")
  case True
  thus ?thesis by simp
next
  case False 
  let ?SR = "S \<union> R"
  from xyL and LSR have "(x,y) \<in> ?SR^*" by simp
  from this and False have "(x,y) \<in> R^* O S O ?SR^*" 
  proof (induct rule: rtrancl_induct)
    case base thus ?case by simp
  next
    case (step y z)
    thus ?case
    proof (cases "(x,y) \<in> R^*")
      case False with step have "(x,y) \<in> R^* O S O ?SR^*" by simp
      from this obtain u where xu: "(x,u) \<in> R^* O S" and uy: "(u,y) \<in> ?SR^*" by force
      from `(y,z) \<in> ?SR` have "(y,z) \<in> ?SR^*" by auto
      with uy have "(u,z) \<in> ?SR^*" by (rule rtrancl_trans)
      with xu show ?thesis by auto
    next
      case True 
      have "(y,z) \<in> S" 
      proof (rule ccontr)
	assume "(y,z) \<notin> S" with `(y,z) \<in> ?SR` have "(y,z) \<in> R" by auto
	with True  have "(x,z) \<in> R^*"  by auto
	with `(x,z) \<notin> R^*` show False ..
      qed
      with True show ?thesis by auto
    qed
  qed
  with LSR show ?thesis by simp
qed  


end
