theory Icap_Impl
imports
  Term_Impl
  Dependency_Graph
  RBT_Map_Set_Extension
begin


(*FIXME: move*)
lemma notin_NF_E[elim]:
  fixes R :: "('f, 'v) trs"
  assumes "t \<notin> NF_trs R"
  obtains C l and \<sigma> :: "('f, 'v) subst" where "l \<in> lhss R" and "t = C\<langle>l \<cdot> \<sigma>\<rangle>"
proof -
  assume 1: "\<And>l C (\<sigma>::('f, 'v) subst). l \<in> lhss R \<Longrightarrow> t = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> thesis"
  from assms obtain u where "(t, u) \<in> rstep R" by (auto simp: NF_def)
  from rstepE[OF this] obtain C l r and \<sigma> :: "('f, 'v) subst" where "(l, r) \<in> R" and "t = C\<langle>l \<cdot> \<sigma>\<rangle>" .
  with 1 show ?thesis by force
qed

lemma NF_subset_conditions:
  fixes R :: "('f, 'v) trs" and Q :: "('f, 'v) terms"
  shows "(\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>) \<longleftrightarrow> NF_terms Q \<subseteq> NF_trs R"
proof (intro iffI allI impI subsetI)
  fix t assume 1: "\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>" and 2: "t \<in> NF_terms Q"
  show "t \<in> NF_trs R"
  proof (rule ccontr)
    assume "t \<notin> NF_trs R"
    then obtain C l and \<sigma> :: "('f, 'v) subst" where "l \<in> lhss R" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" by blast
    with 1[THEN bspec, OF `l \<in> lhss R`]
      obtain u D and \<tau> :: "('f, 'v) subst" where "u \<in> Q" and l: "l = D\<langle>u \<cdot> \<tau>\<rangle>" by auto
    hence "(u, u) \<in> Id_on Q" by auto
    from rstep.ctxt[OF rstep.subst[OF rstep.ctxt[OF rstep.subst[OF rstep.id[OF this]]]], of C D "\<tau>" "\<sigma>",
      unfolded l[symmetric] t[symmetric]]
      show False using `t \<in> NF_terms Q` by (auto simp: NF_def)
  qed
next
  assume 1: "NF_terms Q \<subseteq> NF_trs R"
  show "\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>"
  proof
    fix l assume "l \<in> lhss R"
    then obtain r where "(l, r) \<in> R" by auto
    from rstep.id[OF this] have "l \<notin> NF_trs R" by (auto simp: NF_def)
    with 1 have "l \<notin> NF_terms Q" by blast
    then obtain C l' and \<sigma> :: "('f, 'v) subst" where "l' \<in> Q" and "l = C\<langle>l' \<cdot> \<sigma>\<rangle>" by auto
    thus "\<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>" by auto
  qed
qed

lemma ctxt_supteq_ex: "(\<exists>C \<sigma>. t = C\<langle>u \<cdot> \<sigma>\<rangle>) \<longleftrightarrow> (\<exists>\<sigma>. t \<unrhd> u \<cdot> \<sigma>)"
  using supteq_ctxt_conv[of t] by auto

lemma aux: "(NF_terms Q \<subseteq> NF_trs R) \<longleftrightarrow> (\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>\<sigma>. l \<unrhd> t \<cdot> \<sigma>)"
  using NF_subset_conditions[of R Q] unfolding ctxt_supteq_ex by blast

(*check "NF_terms (set Q) \<subseteq> NF_trs (set R)"*)
definition NF_subset :: "('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> bool" where
  "NF_subset ls Q \<equiv> list_all (\<lambda>l. list_ex (\<lambda>t. list_ex (\<lambda>u. matches u t) (supteq_list l)) Q) ls"

lemma supteq_list_ex: "(\<exists>u\<in>set (supteq_list l). \<exists>\<sigma>. u = t \<cdot> \<sigma>) \<longleftrightarrow> (\<exists>\<sigma>. l \<unrhd> t \<cdot> \<sigma>)" unfolding supteq_list  by auto

lemma NF_subset_sound[simp]:
  "NF_subset (map fst R) Q \<longleftrightarrow> NF_terms (set Q) \<subseteq> NF_trs (set R)"
  using aux[of "set Q" "set R", symmetric]
  unfolding NF_subset_def
  unfolding list_ex_iff
  unfolding list_all_iff
  unfolding matches
  unfolding supteq_list_ex
  unfolding set_map
  .

lemma efficient_NF_check[simp]: "lhss R \<inter> NF_terms Q = {} \<longleftrightarrow> NF_terms Q \<subseteq> NF_trs R"
  using NF_terms_subset_criterion[of "lhss R" Q]
  unfolding NF_terms_lhss .

lemma all_not_conv: "(\<forall>x \<in> A. \<not> P x) = (\<not> (\<exists>x \<in> A. P x))" by auto

subsection {*Efficient Normal Form Checking*}

lemma lhs_var_not_NF:
  assumes "l \<in> T" and "is_var l" shows "t \<notin> NF_terms T"
proof -
  from assms obtain x where l: "l = Var x" sorry
  let ?\<sigma> = "{x/t}"
  from assms have "l \<notin> NF_terms T" by auto
  with NF_instance[of l "?\<sigma>" "Id_on T"]
    have "l \<cdot> ?\<sigma> \<notin> NF_terms T" by auto
  thus ?thesis by (simp add: l subst_def)
qed

lemma neq_root_no_match:
  assumes "\<not> is_var l" and "the (root l) \<noteq> the (root t)"
  shows "\<not> matches t l"
using assms sorry

type_synonym ('f, 'v) term_map = "('f, ('f, 'v) term list) rm"

definition term_map :: "('f::linorder, 'v) term list \<Rightarrow> ('f, 'v) term_map" where
  "term_map ts = elem_list_to_rm (the \<circ> root) ts"

lemma rm_set_lookup_subset:
  "set (rm_set_lookup (term_map ts) x) \<subseteq> set ts"
  using elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ts x]
  by (auto simp: o_def term_map_def)

lemma efficient_ball:
  assumes "\<forall>t\<in>set ts. the (root t) \<noteq> the (root s) \<longrightarrow> P s t"
  shows
    "(\<forall>t\<in>set ts. P s t) \<longleftrightarrow>
     (\<forall>t\<in>set (rm_set_lookup (term_map ts) (the (root s))). P s t)"
  using assms
    and elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ts "the (root s)"]
  by (auto simp: o_def term_map_def)

lemma efficient_supteq_list_do_not_match:
  assumes "\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
  shows
    "(\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). \<not> matches u l) \<longleftrightarrow>
     (\<forall>u\<in>set (supteq_list t). \<forall>l\<in>set (rm_set_lookup (term_map ls) (the (root u))).
       \<not> matches u l)"
    (is "?lhs \<longleftrightarrow> ?rhs" is "_ \<longleftrightarrow> (\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l)")
proof (intro iffI ballI)
  fix u l assume "\<forall>l\<in>set ls. \<forall>u\<in>set ?subs. \<not> matches u l" and "u \<in> set ?subs"
    and "l \<in> set (?ls u)"
  thus "\<not> matches u l"
    using elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ls "the (root u)"]
    by (auto simp: o_def term_map_def)
next
  fix l u assume 1: "\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l"
    and "l \<in> set ls" and "u \<in> set ?subs"
  with assms have "the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
    and "\<forall>l\<in>set (?ls u). \<not> matches u l " by simp+
  with elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ls "the (root u)"]
    and `l \<in> set ls`
    show "\<not> matches u l" by (auto simp: o_def term_map_def)
qed

definition is_NF :: "bool \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f::linorder, 'v) term_map \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "is_NF var_cond R m = (if var_cond then (\<lambda> t. False) 
     else if R = [] then (\<lambda> t. True) 
     else (\<lambda> t. list_all (\<lambda>u. list_all (\<lambda>l. \<not> matches u l)
      (rm_set_lookup m (the (root u)))) (supteq_list t)))"

lemma is_NF_NF_trs_conv:
  "is_NF  (list_ex (is_var \<circ> fst) R) (map fst R) (term_map (map fst R)) t \<longleftrightarrow> t \<in> NF_trs (set R)"
    (is "is_NF ?var ?R ?map t \<longleftrightarrow> _")
proof (intro iffI allI)
  assume is_NF: "is_NF ?var ?R ?map t"
  show "t \<in> NF_trs (set R)"
  proof (cases "list_ex (is_var \<circ> fst) R")
    case True with is_NF[unfolded is_NF_def] show ?thesis by simp
  next
    case False    
    hence novars: "list_ex (is_var \<circ> fst) R = False" by auto
    let ?ts = "map fst R"
    from False have "\<forall>s\<in>set ?ts. \<not> is_var s" by (simp add: list_ex_iff o_def)
    with neq_root_no_match
    have no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
        \<longrightarrow> \<not> matches u s" by blast
    note is_NF = is_NF[unfolded is_NF_def novars if_False list_all_iff]
    show ?thesis
    proof (cases "(R = []) = False")
      case True
      from is_NF[simplified, unfolded True if_False]
      show ?thesis
        unfolding efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches
        unfolding supteq_list_ex by auto
    next
      case False
      hence "set R = {}" by auto
      thus ?thesis by auto
    qed
  qed
next
  assume NF_trs: "t \<in> NF_trs (set R)"
  show "is_NF (list_ex (is_var \<circ> fst) R) (map fst R) (term_map (map fst R)) t"
  proof (cases "list_ex (is_var \<circ> fst) R")
    case True
    then obtain l where "l \<in> lhss (set R)" and "is_var l"
      unfolding list_ex_iff o_def by auto
    from lhs_var_not_NF[OF this] and NF_trs show ?thesis by simp
  next
    case False
    hence novars: "list_ex (is_var \<circ> fst) R = False" by simp
    let ?ts = "map fst R"
    from False have "\<forall>s\<in>set ?ts. \<not> is_var s" by (simp add: list_ex_iff o_def)
    with neq_root_no_match
    have
      no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
      \<longrightarrow> \<not> matches u s" by blast
    show ?thesis
    proof (cases "(map fst R = []) = False")
      case False thus ?thesis unfolding is_NF_def novars by auto
    next
      case True
      from NF_trs
      show ?thesis
        unfolding is_NF_def novars True if_False list_all_iff
        unfolding efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches
        unfolding supteq_list_ex set_map by force
    qed
  qed
qed

lemma is_NF_NF_terms_conv:
  "is_NF (list_ex is_var T) T (term_map T) t \<longleftrightarrow> t \<in> NF_terms (set T)"
proof -
  let ?T = "map (\<lambda>t. (t, t)) T"
  have 1: "list_ex is_var T = list_ex (is_var \<circ> fst) ?T"
    by (induct T) (auto simp: o_def)
  have 2: "map fst ?T = T" by (induct T, auto)
  have 3: "term_map T = term_map (map fst ?T)" unfolding 2 ..
  have 4: "set ?T = Id_on (set T)" by (induct T) (auto simp: o_def)
  from is_NF_NF_trs_conv[of "?T" t]
    show ?thesis unfolding 1 3 4 unfolding 2 .
qed

fun icap_impl ::
  "bool \<Rightarrow> (('f, 'v + int + 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term list
    \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('v \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, unit + 'v) term"
where
  "icap_impl nf isQnf ls S Sx (Var x) =
    (if nf \<and> Sx x then Var (Inr x) else Var (Inl ()))"
| "icap_impl nf isQnf ls S Sx (Fun f ts) = (
    let t' = Fun f (map (icap_impl nf isQnf ls S Sx) ts) in
    if list_ex (\<lambda>l. check_class (\<lambda>\<mu>. list_all (\<lambda>u. isQnf (u \<cdot> \<mu>)) (args l))
      (\<lambda>\<mu>. list_all (\<lambda>u. isQnf (u \<cdot> map_subst_domain Inr \<mu>)) S) l t') ls
      then Var (Inl ())
      else t')"

text{*The efficient implementation of @{const icap} is correct.*}
lemma icap_impl_sound[simp]:
  fixes Q :: "('f::linorder, 'v) term list" and S
  defines InlQ: "Q' \<equiv> map (map_vars Inl) Q"
      and Sx: "Sx \<equiv> \<lambda>x. (\<exists> t \<in> set S. x \<in> vars_term t)"
  shows
  "icap_impl (NF_subset (map fst R) Q)
             (is_NF (list_ex is_var Q') Q' (term_map Q'))
             (map fst R) S Sx t =
    icap (set R) (set Q) (set S) t"
proof (induct t)
  case (Var x) show ?case by (simp add: Sx)
next
  case (Fun f ts)
  hence IH: "map (icap_impl (NF_subset (map fst R) Q)
                            (is_NF (list_ex is_var Q') Q' (term_map Q'))
                            (map fst R) S Sx) ts =
    map (icap (set R) (set Q) (set S)) ts" by auto
 show ?case
    unfolding icap_impl.simps
    unfolding IH
    unfolding is_NF_NF_terms_conv
    using icap.simps(2)[of "set R" "set Q" "set S" f ts]
    unfolding list_all_iff
    unfolding list_ex_iff
    unfolding InlQ
    unfolding set_map
    by simp
qed

definition icap_impl' :: "_" where
  "icap_impl' R Q S t = (
    let ls = map fst R;
        Q' = map (map_vars Inl) Q;
        isnf = is_NF (list_ex is_var Q') Q' (term_map Q');
        nf = NF_subset ls Q;
        Sxs = list_to_rs (concat (map vars_term_list S));
        Sx = (\<lambda> x. rs_memb x Sxs) in
    icap_impl nf isnf ls S Sx t)"

lemma icap_impl'_sound[simp]: "icap_impl' R Q S t = icap (set R) (set Q) (set S) t"
  unfolding icap_impl'_def
  unfolding Let_def
  unfolding icap_impl_sound[symmetric] 
  by (simp add: rs_correct) 

end