header{*Multiset extension of the (non-)strict order*}

theory Multiset_Extension
imports Multiset2

begin

subsection{* Definition of the orders*}

type_synonym 'a relation = "'a \<Rightarrow> 'a \<Rightarrow> bool"

definition ns_mul_ext :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> ('a multiset) relation"
 where "ns_mul_ext ns s A B \<equiv> \<exists> as bs A' B'. (
  A = multiset_of as + A' \<and>
  B = multiset_of bs + B' \<and>
  length as = length bs \<and>
  (\<forall>i. i < length as \<longrightarrow> ns (as ! i) (bs ! i)) \<and>
  (\<forall>a. a \<in># A' \<longrightarrow> (\<exists>b. b \<in># B' \<and> s a b))
  )"

definition s_mul_ext :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> ('a multiset) relation"
 where "s_mul_ext ns s A B \<equiv> \<exists> as bs A' B'. (
  A = multiset_of as + A' \<and>
  B = multiset_of bs + B' \<and>
  length as = length bs \<and>
  B' \<noteq> {#} \<and>
  (\<forall>i. i < length as \<longrightarrow> ns (as ! i) (bs ! i)) \<and>
  (\<forall>a. a \<in># A' \<longrightarrow> (\<exists>b. b \<in># B' \<and> s a b))
  )"

declare ns_mul_ext_def[simp]
declare s_mul_ext_def[simp]

subsection{* Basic properties *}

text{* The empty multiset is the minimal element for these orders *}

lemma ns_mul_ext_bottom : "ns_mul_ext ns s {#} A"
by auto

lemma s_mul_ext_bottom : assumes "A \<noteq> {#}"
  shows "s_mul_ext ns s {#} A"
using assms by auto

text{* Being stricly lesser than implies being lesser than*}

lemma s_ns_mul_ext : "s_mul_ext ns s A B \<longrightarrow> ns_mul_ext ns s A B"
by auto

text{* In a multiset, there exists a local maxima. *}

lemma multiset_of_list_of : "multiset_of (list_of A) = A"
by auto

lemma Melem_multiset_of_step :
  assumes "a \<in># multiset_of xs"
  shows "a \<in># multiset_of (x # xs)"
using assms by auto

lemma list_empty_inv :
  assumes "list_of A = []"
  shows "A = {#}"
proof-
  from assms and multiset_of_list_of[of A] show ?thesis by auto
qed

lemma local_maxima_multiset_aux :
  assumes s_trans: "\<And> x y z. s x y \<longrightarrow> s y z \<longrightarrow> s x z"
  and s_irrefl: "\<And>x. \<not> s x x"
  shows "xs \<noteq> [] \<longrightarrow> (\<exists>a. a \<in># (multiset_of xs) \<and> (\<forall>b. b \<in># (multiset_of xs) \<longrightarrow> \<not> s a b))"
proof (induct xs)
case Nil
 show ?case by auto
next
case (Cons x xs)
 assume IH: "xs \<noteq> [] \<longrightarrow> (\<exists>a. a \<in># multiset_of xs \<and> (\<forall>b. b \<in># multiset_of xs \<longrightarrow> \<not> s a b))"
 show ?case
  proof (cases xs)
  case Nil
   from s_irrefl have "x \<in># multiset_of [x] \<and> (\<forall>b. b \<in># multiset_of [x] \<longrightarrow> \<not> s x b)" by simp
   from this and Nil[symmetric]
   have "x \<in># multiset_of (x # xs) \<and> (\<forall>b. b \<in># multiset_of (x # xs) \<longrightarrow> \<not> s x b)" by auto
   thus ?thesis by blast
  next
  case (Cons y ys)
   from this and IH obtain a1 where
    a1_in_xs: "a1 \<in># multiset_of xs"
    and a1_is_lm: "\<forall>b. b \<in># multiset_of xs \<longrightarrow> \<not> s a1 b" by auto
   show ?thesis 
    proof (cases "s a1 x")
    case True
     assume sa1x: "s a1 x"
     have GL: "x \<in># multiset_of (x # xs)" by auto
     from s_irrefl have GR1: "\<not> s x x" by simp
     {
       fix b
       have "b \<in># multiset_of xs \<longrightarrow> \<not> s x b"
       proof (cases "s x b")
        case True
         from this and sa1x and s_trans have "s a1 b" by blast
         from this and a1_is_lm show ?thesis by auto
        next
        case False
         thus ?thesis by auto
        qed
     }
     hence GR2: "\<forall>b. b \<in># multiset_of xs \<longrightarrow> \<not> s x b" by auto
     from this and GL and GR1 show ?thesis by auto
    next
    case False
     from this and a1_is_lm
     have GR: "\<forall>b. b \<in># multiset_of (x # xs) \<longrightarrow> \<not> s a1 b" by auto
     from a1_in_xs and Melem_multiset_of_step have GL: "a1 \<in># multiset_of (x # xs)" by auto
     from GL and GR show ?thesis by blast
    qed
  qed
qed

lemma local_maxima_multiset :
  assumes s_trans: "\<And> x y z. s x y \<longrightarrow> s y z \<longrightarrow> s x z"
  and s_irrefl: "\<And>x. \<not> s x x"
  and A_ne: "A \<noteq> {#}"
  shows "\<exists>a. a \<in># A \<and> (\<forall>b. b \<in># A \<longrightarrow> \<not> s a b)"
proof-
 from A_ne have lA_ne: "list_of A \<noteq> []"
  proof (cases "list_of A")
   case Nil
    from multiset_of_list_of[symmetric, of A] and this have "A = {#}" by auto
    from this and A_ne show ?thesis by auto
   next
   case Cons thus ?thesis by auto
  qed
 from this and assms and local_maxima_multiset_aux
 have "\<exists>a. a \<in># (multiset_of (list_of A)) \<and> (\<forall>b. b \<in># (multiset_of (list_of A)) \<longrightarrow> \<not> s a b)"
 by blast
 from this and multiset_of_list_of show ?thesis by simp
qed

text{* The non-strict order is reflexive *}

lemma ns_mul_ext_refl : assumes ns_refl: "\<And>x. ns x x"
  shows "ns_mul_ext ns s A A"
proof -
 from ns_refl have AnsA: "\<forall>i. i < length (list_of A) \<longrightarrow> ns (list_of A ! i) (list_of A ! i)" 
 by auto
 from multiset_of_list_of[symmetric] and empty_neutral[symmetric]
 have "A = multiset_of (list_of A) + {#}" by auto
 from this and AnsA
 have "A = multiset_of (list_of A) + {#} \<and>
       A = multiset_of (list_of A) + {#} \<and>
       length (list_of A) = length (list_of A) \<and>
       (\<forall>i. i < length (list_of A) \<longrightarrow> ns (list_of A ! i) (list_of A ! i)) \<and>
       (\<forall>a. a \<in># {#} \<longrightarrow> (\<exists>b. b \<in># {#} \<and> s a b))" by auto
 thus ?thesis unfolding ns_mul_ext_def by blast
qed

text{* The non-strict order is transitive *}

lemma ns_mul_ext_trans :
  assumes ns_trans: "\<And> x y z. ns x y \<longrightarrow> ns y z \<longrightarrow> ns x z"
  and s_trans: "\<And> x y z. s x y \<longrightarrow> s y z \<longrightarrow> s x z"
  and s_ns_compat: "\<And> x y z. s x y \<longrightarrow> ns y z \<longrightarrow> s x z"
  and ns_s_compat: "\<And> x y z. ns x y \<longrightarrow> s y z \<longrightarrow> s x z"
  and AnsB: "ns_mul_ext ns s A B"
  and BnsC: "ns_mul_ext ns s B C"
  shows "A = B"
proof-
(* sketch of the proof:
    A = {as} + A' and B = {b1s} + B1 from AnsB
    B = {b2s} + B2 and C = {cs} + C' from BnsC

    compute as' and cs' the lists of elements
    that are in relation with the one present
    in bs = list_of ({b1s} \<inter># {b2s}) and order
    them properly

    then A = {as'} + (A - {as'}) and
    C = {cs'} + (C - {cs'}) and from this
    prove that A \<le># C
*)
oops

text{* The non-strict order is antisymmetric*}

lemma ns_mul_ext_antisym :
  assumes ns_antisym: "\<And> x y. ns x y \<longrightarrow> ns y x \<longrightarrow> x = y"
  and s_irrefl: "\<And> x. \<not> s x x"
  and s_trans: "\<And> x y z. s x y \<longrightarrow> s y z \<longrightarrow> s x z"
  and AnsB: "ns_mul_ext ns s A B"
  and BnsA: "ns_mul_ext ns s B A"
  shows "A = B"c
proof-
(* sketch of the proof:
    suppose A \<le># B (A = {as} + A') and B \<le> A
    pose a\<top> = local maxima of A

    ** if a\<top> is in {as} then:
       \<exists>b \<in># B such that a\<top> \<le> b
       then \<exists>a' \<in># A such that b \<le> a
       therefore a\<top> \<le> a'
         if a\<top> < a then contradiction
         if a\<top> = a then a\<top> = b and we can use the IH

    ** if a\<top> is in A' then
       \<exists>b \<in># B such that a\<top> < b
       then \<exists>a' \<in># A such that b \<le> a'
       therefore a\<top> < a' i.e. contradiction
*)
oops