(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Poly_Order
imports Term_Order Polynomial SN_Orders2
begin

type_synonym ('f, 'a) poly_inter = "'f \<Rightarrow> nat \<Rightarrow> (nat, 'a) poly"

fun eval_term :: "('f,'a :: poly_carrier)poly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)poly"
where "eval_term _ (Var x) = [([(x,1)],1)]"
  |   "eval_term I (Fun f ts) = (let ps = map (eval_term I) ts; n = length ts in poly_subst (\<lambda> i. if i < n then ps ! i else zero_poly) (I f n))"


locale poly_order = poly_order_carrier +
  fixes I :: "('f,'a)poly_inter"
    and type :: 'v
    and \<pi> :: "'f af"
  assumes  pos_I: "\<And> f n. I f n \<ge>p zero_poly"
  and mono_I: "\<And> f n. poly_weak_mono_all type (I f n)" 
  and pi: "\<And> f n i. \<not> (\<pi> f n i) \<Longrightarrow> \<not> (poly_var (I f n) i)"
begin

lemma eval_term_pos: fixes t :: "('f,'v)term" shows "eval_term I t \<ge>p zero_poly"
unfolding poly_ge_def zero_poly_def
proof (intro impI allI)
  fix \<alpha> :: "('v,'a)assign"
  assume pos: "pos_assign \<alpha>"
  show "eval_poly \<alpha> (eval_term I t) \<ge> eval_poly \<alpha> []"
  proof (induct t)
    case (Var x) thus ?case by (simp add: pos[unfolded pos_assign_def])
  next
    case (Fun f ts)
    {
      fix i
      have "eval_poly \<alpha> (if i < length ts then map (eval_term I) ts ! i else zero_poly) \<ge> 0"
      proof (cases "i < length ts")
        case False thus ?thesis unfolding zero_poly_def by (simp add: ge_refl)
      next
        case True
        hence "ts ! i \<in> set ts" by auto
        with Fun have "eval_poly \<alpha> (eval_term I (ts ! i)) \<ge> 0" by auto
        with True show ?thesis by (simp)
      qed
    }
    thus ?case 
      by (simp add: Let_def poly_subst, intro pos_assign_poly[OF _ pos_I], unfold pos_assign_def, auto)
  qed
qed

definition inter_s :: "('f,'v)trs"
  where "inter_s \<equiv> {(s,t). (eval_term I s >p eval_term I t)}"

definition inter_ns :: "('f,'v)trs"
  where "inter_ns \<equiv> {(s,t). eval_term I s \<ge>p eval_term I t}"

lemma inter_stable: 
  shows "eval_poly \<alpha> (eval_term I (t \<cdot> \<sigma>)) = eval_poly (\<lambda> x. eval_poly \<alpha> (eval_term I (\<sigma> x))) (eval_term I t)"
proof -
  have "eval_poly \<alpha> (eval_term I (t \<cdot> \<sigma>)) = eval_poly (\<lambda> x. eval_poly \<alpha> (eval_term I (\<sigma> x))) (eval_term I t)" (is "_ = eval_poly ?ass _")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ts)
    let ?ts = "map (\<lambda> s. s \<cdot> \<sigma>) ts"
    show ?case
    proof (simp add: Let_def poly_subst, rule fun_cong, rule arg_cong[where f = eval_poly], rule ext, unfold zero_poly_def)
      fix i
      show "eval_poly \<alpha> (if i < length ts then map (eval_term I \<circ> (\<lambda> t. t \<cdot> \<sigma>)) ts ! i else []) = 
            eval_poly ?ass (if i < length ts then map (eval_term I) ts ! i else [])"
      proof (cases "i < length ts")
        case False thus ?thesis by simp
      next
        case True
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] True show ?thesis by simp
      qed
    qed
  qed
  thus ?thesis by simp
qed

end

sublocale poly_order \<subseteq> af_redpair inter_s inter_ns \<pi>
proof
  show "subst_closed inter_s"
    unfolding subst_closed_def
  proof
    fix ss ts :: "('f,'v)term"
    assume "(ss,ts) \<in> CS inter_s"
    thus "(ss,ts) \<in> inter_s"
    proof
      fix s t sig
      assume ss: "ss = s \<cdot> sig" and ts: "ts = t \<cdot> sig" and st: "(s,t) \<in> inter_s"
      thus ?thesis using eval_term_pos[unfolded poly_ge_def zero_poly_def] 
        by (force simp: inter_s_def poly_gt_def inter_stable pos_assign_def eval_term_pos)
    qed
  qed
next
  show "subst_closed inter_ns"
    unfolding subst_closed_def
  proof
    fix ss ts :: "('f,'v)term"
    assume "(ss,ts) \<in> CS inter_ns"
    thus "(ss,ts) \<in> inter_ns"
    proof
      fix s t sig
      assume ss: "ss = s \<cdot> sig" and ts: "ts = t \<cdot> sig" and st: "(s,t) \<in> inter_ns"
      thus ?thesis using eval_term_pos[unfolded poly_ge_def zero_poly_def] 
        by (force simp: inter_ns_def poly_ge_def inter_stable pos_assign_def eval_term_pos)
    qed
  qed
next
  show "SN inter_s"
  proof
    fix f
    assume steps: "\<forall> i. (f i, f (Suc i)) \<in> inter_s"
    obtain g where g: "\<And> i. g i = eval_term I (f i)" by auto
    from steps[unfolded inter_s_def] eval_term_pos
    have "\<And> i. (g i, g (Suc i)) \<in> poly_GT" 
      by (simp only: g, auto)
    with poly_GT_SN show False unfolding SN_defs by blast
  qed
next
  show "inter_ns O inter_s \<subseteq> inter_s"
    by (rule subsetI2, clarify, unfold inter_s_def inter_ns_def, simp add: poly_compat)
next
  show"one_step_ctxt_closed inter_ns"
    unfolding one_step_ctxt_closed_def
  proof (intro allI impI)
    fix f bef and s t :: "('f,'v)term" and aft 
    assume st: "(s,t) \<in> inter_ns"
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_ns" (is "(Fun f ?s, Fun f ?t) \<in> _")
      unfolding inter_ns_def poly_ge_def
    proof (clarify)
      fix \<alpha> :: "('v,'a)assign"
      assume pos: "pos_assign \<alpha>"
      let ?n = "Suc (length bef + length aft)"
      let ?i = "length bef"
      from mono_I[of f ?n] have mono: "poly_weak_mono type (I f ?n) ?i" by (rule poly_weak_mono_all)
      let ?exp = "\<lambda> w s.  (if w < Suc (length bef + length aft) then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! w else zero_poly)"
      {
        fix w
        assume "?i \<noteq> w"
        hence "?exp w s = ?exp w t" by (simp add: nth_append, cases "w - ?i", auto)
      } note one = this
      {
        fix w
        have "\<exists> ts. (?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly))"
          by (rule exI[of _ ?t], simp only: map_append, simp)
        then obtain ts where id: "?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly)" by blast
        have "?exp w t \<ge>p zero_poly" 
          by (simp only: id, unfold poly_ge_def zero_poly_def, simp add: ge_refl, intro impI allI, force simp: eval_term_pos[unfolded poly_ge_def zero_poly_def, simplified])
      } note two = this 
      have "eval_poly \<alpha> (poly_subst (\<lambda>i. if i < ?n then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! i else zero_poly) (I f ?n)) \<ge>
        eval_poly \<alpha> (poly_subst (\<lambda>i. if i < ?n then (map (eval_term I) bef @ eval_term I t # map (eval_term I) aft) ! i else zero_poly) (I f ?n))"
        by (rule mono[unfolded poly_weak_mono_def poly_ge_def, THEN spec, THEN spec, THEN mp, THEN spec, THEN mp[OF _ pos]], rule conjI, 
        (auto simp: one two ge_refl two[unfolded poly_ge_def])[1], simp add: nth_append st[unfolded inter_ns_def poly_ge_def, simplified])
      thus "ge (eval_poly \<alpha> (eval_term I (Fun f ?s))) (eval_poly \<alpha> (eval_term I (Fun f ?t)))" by simp
    qed
  qed
next
  show "af_compatible \<pi> inter_ns"
    unfolding af_compatible_def
  proof (intro allI, clarify)
    fix f bef s t aft
    assume ns: "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<notin> inter_ns"
    let ?s = "bef @ s # aft"
    let ?t = "bef @ t # aft"
    let ?n = "Suc (length bef + length aft)"
    let ?i = "length bef"
    show "\<pi> f ?n ?i"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      with pi have pv: "\<not> poly_var (I f ?n) ?i" by simp
      let ?sub = "\<lambda>s i. if i < ?n then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! i else zero_poly"
      have id: "poly_subst (?sub s) (I f ?n) = poly_subst (?sub t) (I f ?n)"
      proof (rule poly_var[OF pv])
        fix j
        assume "?i \<noteq> j"
        thus "(if j < ?n then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! j else zero_poly) = 
          (if j < ?n then (map (eval_term I) bef @ eval_term I t # map (eval_term I) aft) ! j else zero_poly)"
          by (cases "j \<ge> ?n", simp, cases "j < ?i", simp add: nth_append, simp, cases "j \<ge> ?i", simp add: nth_append, cases "j - ?i", simp, simp add: nth_append, simp)
      qed
      { 
        fix \<alpha>
        have "eval_poly \<alpha> (eval_term I (Fun f ?s)) = eval_poly \<alpha> (eval_term I (Fun f ?t))"  using id by simp
        hence "eval_poly \<alpha> (eval_term I (Fun f ?s)) \<ge> eval_poly \<alpha> (eval_term I (Fun f ?t))" using ge_refl by simp
      }
      hence "(Fun f ?s, Fun f ?t) \<in> inter_ns" unfolding inter_ns_def poly_ge_def by auto
      with ns show False ..
    qed
  qed
qed

locale mono_poly_order = poly_order +
  assumes strict_mono_I: "\<And> f n i. i < n \<Longrightarrow> poly_strict_mono type (I f n) i"
begin

lemma inter_s_mono: "one_step_ctxt_closed inter_s"
unfolding one_step_ctxt_closed_def
proof (intro allI impI)
  fix f bef and s t :: "('b,'c)term" and aft 
  assume st: "(s,t) \<in> inter_s"
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s" (is "(Fun f ?s, Fun f ?t) \<in> _")
    unfolding inter_s_def poly_gt_def
  proof (clarify)
    fix \<alpha> :: "('c,'a)assign"
    assume pos: "pos_assign \<alpha>"
    let ?n = "Suc (length bef + length aft)"
    let ?i = "length bef"
    from strict_mono_I[of ?i ?n] have mono: "poly_strict_mono type (I f ?n) ?i" by simp
    let ?exp = "\<lambda> w s.  (if w < Suc (length bef + length aft) then (map (eval_term I) bef @ eval_term I s # map (eval_term I) aft) ! w else zero_poly)"
    {
      fix w
      assume "?i \<noteq> w"
      hence "?exp w s = ?exp w t" by (simp add: nth_append, cases "w - ?i", auto)
    } note one = this
    {
      fix w
      have "\<exists> ts. (?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly))"
        by (rule exI[of _ ?t], simp only: map_append, simp)
      then obtain ts where id: "?exp w t = (if w < length ts then (map (eval_term I) ts) ! w else zero_poly)" (is "_ = ?if") by blast
      have "?exp w t \<ge>p zero_poly" 
      proof (simp only: id, unfold poly_ge_def, intro impI allI)
        fix \<alpha> :: "('c,'a)assign"
        assume "pos_assign \<alpha>"
        thus "eval_poly \<alpha> ?if \<ge> eval_poly \<alpha> zero_poly"
          by (simp add: ge_refl zero_poly_def, force simp: eval_term_pos[unfolded poly_ge_def zero_poly_def, simplified])
      qed
    } note two = this 
    have "gt (eval_poly \<alpha> (poly_subst (\<lambda> i. ?exp i s) (I f ?n))) (eval_poly \<alpha> (poly_subst (\<lambda> i. ?exp i t) (I f ?n)))"
      by (rule mono[unfolded poly_strict_mono_def poly_gt_def, THEN spec, THEN spec, THEN mp, THEN spec, THEN mp[OF _ pos]], rule conjI, (auto simp: one two)[1], simp add: nth_append st[unfolded inter_s_def poly_gt_def, simplified])
    thus "gt (eval_poly \<alpha> (eval_term I (Fun f ?s))) (eval_poly \<alpha> (eval_term I (Fun f ?t)))" by simp
  qed
qed
end       

definition default_I :: "'a \<Rightarrow> nat \<Rightarrow> (nat,'a :: poly_carrier)poly"
where "default_I def n \<equiv> ([],def) # map (\<lambda> i. ([(i,1)],1)) [0 ..< n]"

lemma default_I_check_poly_weak_mono: assumes "ge def (0 :: 'a :: poly_carrier)" shows "check_poly_weak_mono_all (default_I def n)"
unfolding default_I_def check_poly_weak_mono_all_def
by (simp add: list_all_iff assms one_ge_zero)

lemma default_I_check_poly_strict_mono: assumes i: "i < n" shows "check_poly_strict_mono power_mono (default_I def n) i"
unfolding check_poly_strict_mono_def list_ex_iff default_I_def using i
by (auto simp: ge_refl check_monom_strict_mono_def)


lemma default_I_pos: assumes "ge def (0 :: 'a :: poly_carrier)" shows "default_I def n \<ge>p zero_poly"
using check_poly_weak_mono_all_pos[OF default_I_check_poly_weak_mono[OF assms]] .

lemma default_I_mono_all: assumes ge: "ge def (0 :: 'a :: poly_carrier)" shows "poly_weak_mono_all type (default_I def n)"
using check_poly_weak_mono_all[OF default_I_check_poly_weak_mono[OF assms]] .

locale ce_poly_order = poly_order +
  assumes default: "\<exists> n. \<forall> f m. m \<ge> n \<longrightarrow> I f m = default_I default m"

sublocale ce_poly_order \<subseteq> mono_ce_af_redpair inter_s inter_ns \<pi>
proof -
  have "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns \<inter> inter_s"
  proof -
    from default obtain n where default: "\<And> f m. m \<ge> n \<Longrightarrow> I f m = default_I default m" by auto
    show ?thesis
    proof(rule exI[of _ n], intro allI impI)
      fix m and c d :: 'b
      assume "n \<le> m"
      with default have default: "I c (Suc (Suc m)) = default_I default (Suc (Suc m))" by auto
      { 
        fix s t :: "('b,'c)term" and \<alpha> :: "('c,'a)assign"
        assume pos: "pos_assign \<alpha>"
        let ?t = "eval_poly \<alpha> (eval_term I t)"
        let ?s = "eval_poly \<alpha> (eval_term I s)"
        from eval_term_pos[of t] eval_term_pos[of s] have s: "?s \<ge> 0" and t: "?t \<ge> 0" using pos unfolding zero_poly_def poly_ge_def by auto
        let ?left = "eval_poly \<alpha> (eval_term I (Fun c (s # (t # replicate m (Fun d [])))))"
        let ?dst = "default + ?s + ?t"
        have main: "?left \<ge> ?dst" 
        proof -
          let ?ass = "(\<lambda>v. eval_poly \<alpha> (if v < Suc (Suc m) then (eval_term I s # eval_term I t # replicate m (poly_subst (\<lambda>i. zero_poly) (I d 0))) ! v else zero_poly))"
          let ?ass2 = "(\<lambda>v. eval_poly \<alpha> (if v < Suc (Suc m) then (eval_term I s # eval_term I t # replicate m (eval_term I (Fun d []))) ! v else zero_poly))"
          let ?f = "(\<lambda>i :: nat. ([(i, 1 :: nat)], 1\<Colon>'a))"
          let ?is = "[Suc (Suc 0)..<Suc (Suc m)]"
          obtain ass iis where ass: "ass = ?ass" and iis: "iis = ?is" by auto
          hence "eval_poly ?ass (([],default) # map ?f (0 # Suc 0 # ?is)) = 
                 eval_poly ass (([],default) # map ?f (0 # Suc 0 # iis))" (is "?left = _") by simp
          also have "\<dots> = ?dst + eval_poly ass (map ?f iis)" by (auto simp: field_simps ass)
          also have "\<dots> \<ge> ?dst"
          proof -
            have "eval_poly ass (map ?f iis) \<ge> 0" 
            proof (rule check_poly_weak_mono_all_pos[unfolded poly_ge_def zero_poly_def, THEN spec[of _ ass], THEN mp, simplified])
              show "check_poly_weak_mono_all (map ?f iis)" unfolding check_poly_weak_mono_all_def list_all_iff 
                using one_ge_zero by auto
            next
              show "pos_assign ass" unfolding pos_assign_def
              proof
                fix i
                show "ass i \<ge> 0"
                proof (cases "i = 0")
                  case True 
                  thus ?thesis by (simp add: s ass)
                next 
                  case False 
                  show ?thesis 
                  proof (cases "i = 1")
                    case True 
                    thus ?thesis by (simp add: t ass)
                  next
                    case False
                    show ?thesis 
                    proof (cases "i \<ge> Suc (Suc m)")
                      case True 
                      thus ?thesis by (simp add: ass zero_poly_def ge_refl)
                    next
                      case False
                      from `i \<noteq> 0` and `i \<noteq> 1` and `\<not> Suc (Suc m) \<le> i`
                      have "i \<ge> 2" and "i < Suc (Suc m)" by auto
                      then obtain j where i: "i = Suc (Suc j)" and j: "j < m" by (cases i, simp, cases "i - 1", auto)
                      show "ass i \<ge> 0" unfolding i ass 
                      proof (simp add: j)
                        have "eval_poly \<alpha> (poly_subst (\<lambda> i. zero_poly) (I d 0)) = eval_poly \<alpha> (eval_term I (Fun d []))" (is "?left = _")
                          by simp
                        also have "\<dots> \<ge> 0" using eval_term_pos[unfolded poly_ge_def zero_poly_def, of "Fun d []"] pos by auto
                        finally show "?left \<ge> 0" .
                      qed
                    qed
                  qed
                qed
              qed
            qed
            from plus_left_mono[OF this, of ?dst] show ?thesis by (auto simp: field_simps)
          qed
          finally show ?thesis by (simp add: default poly_subst, simp only: default_I_def upt_conv_Cons)
        qed
        from main have mains: "?left \<ge> ?dst" .
        also have "\<dots> \<ge> ?s + ?t" using plus_left_mono[OF default_ge_zero, of "?s + ?t"] by (auto simp: field_simps)
        finally have main: "?left \<ge> ?s + ?t" .
        from main have "?left \<ge> ?s + ?t" .
        also have "\<dots> \<ge> ?s" using plus_left_mono[OF t, of ?s] by (auto simp: field_simps)
        finally have sge: "?left \<ge> ?s" .
        from main have "?left \<ge> ?s + ?t" .
        also have "\<dots> \<ge> ?t" using plus_left_mono[OF s, of ?t] by (auto simp: field_simps)
        finally have tge: "?left \<ge> ?t" .
        from mains have "?left \<ge> ?dst" .
        also have "\<dots> \<ge> default + ?s" using plus_left_mono[OF t, of "default + ?s"] by (auto simp: field_simps)
        also have "gt \<dots> ?s" using plus_gt_left_mono[OF default_gt_zero, of ?s] by simp
        finally have sgt: "gt ?left ?s" .
        from mains have "?left \<ge> ?dst" .
        also have "\<dots> \<ge> default + ?t" using plus_left_mono[OF s, of "default + ?t"] by (auto simp: field_simps)
        also have "gt \<dots> ?t" using plus_gt_left_mono[OF default_gt_zero, of ?t] by simp
        finally have tgt: "gt ?left ?t" .
        from sge tge sgt tgt have "?left \<ge> ?s" and "?left \<ge> ?t" and "gt ?left ?s" and "gt ?left ?t" by blast+
      } note main = this
      show "ce_trs (c,d,m) \<subseteq> inter_ns \<inter> inter_s"
        unfolding ce_trs.simps inter_ns_def inter_s_def
        by (auto simp del: eval_term.simps simp: poly_ge_def poly_gt_def main)
    qed
  qed 
  hence one: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_s" and
    two: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns" by blast+
  show "mono_ce_af_redpair inter_s inter_ns \<pi>"
    by (unfold_locales, rule two, rule one)
qed

type_synonym ('f,'a)poly_inter_list = "(('f \<times> nat) \<times> (nat,'a)poly)list"

definition poly_inter_list_to_inter :: "'a :: poly_carrier \<Rightarrow> ('f,'a)poly_inter_list \<Rightarrow> ('f,'a)poly_inter"
where "poly_inter_list_to_inter def I \<equiv> \<lambda> f n. case lookup (f,n) I of None \<Rightarrow> default_I def n | Some p \<Rightarrow> p"

definition poly_inter_to_af :: "('f,'a)poly_inter \<Rightarrow> 'f af"
where "poly_inter_to_af I f n \<equiv> poly_var (I f n)"

definition check_poly_inter_list :: "bool \<Rightarrow> ('f,'a :: poly_carrier)poly_inter_list \<Rightarrow> ('f \<times> (nat,'a)poly) check"
where "check_poly_inter_list discrete I \<equiv> check_all (\<lambda> (_,p). check_poly_weak_mono_and_pos discrete p) I <+? (\<lambda> ((f,n),p). (f,p))"

context poly_order_carrier
begin

lemma check_poly_inter_list: assumes check: "isOK(check_poly_inter_list discrete I)"
  shows "poly_weak_mono_all type (poly_inter_list_to_inter default I f n) \<and> (poly_inter_list_to_inter default I f n \<ge>p zero_poly) "
proof -
  let ?I = "poly_inter_list_to_inter default I"
  show ?thesis
  proof (cases "lookup (f,n) I")
    case None thus ?thesis 
      by (simp only: poly_inter_list_to_inter_def, simp add: default_I_pos[OF default_ge_zero] default_I_mono_all[OF default_ge_zero])
  next 
    case (Some p)
    with lookup_sound[OF this] check[unfolded check_poly_inter_list_def]
    show ?thesis unfolding poly_inter_list_to_inter_def using check_poly_weak_mono_and_pos  by force
  qed
qed

    

lemma ce_poly_order: 
  fixes type :: 'v and I :: "('f,'a)poly_inter_list"
  assumes check: "isOK(check_poly_inter_list discrete I)"
  shows "ce_poly_order default gt power_mono discrete (poly_inter_list_to_inter default I) type (poly_inter_to_af (poly_inter_list_to_inter default I))"
proof -
  let ?I = "poly_inter_list_to_inter default I :: ('f,'a)poly_inter"
  from check_poly_inter_list[OF check, of type] have main1: "\<And> f n. poly_weak_mono_all type (?I f n)" and main2: "\<And> f n. (?I f n) \<ge>p zero_poly" by auto
  show ?thesis
  proof (unfold_locales, rule main2, rule main1)
    let ?m = "Suc (max_list (map (snd o fst) I))"
    show "\<exists> n. \<forall> f m. n \<le> m \<longrightarrow> ?I f m = default_I default m"
    proof (rule exI[of _ ?m], intro allI impI)
      fix f m 
      assume m: "?m \<le> m"
      show "?I f m = default_I default m" 
      proof (cases "lookup (f,m) I")
        case None thus ?thesis 
          by (simp add: poly_inter_list_to_inter_def)
      next
        case (Some p)
        from lookup_sound[OF this] have "((f,m),p) \<in> set I" .
        hence "m \<in> set (map (snd o fst) I)" by force
        from max_list[OF this] m show ?thesis by auto
      qed
    qed
  next
    fix f n i
    assume "\<not> poly_inter_to_af (poly_inter_list_to_inter default I) f n i"
    thus "\<not> poly_var (poly_inter_list_to_inter default I f n) i"
      unfolding poly_inter_to_af_def .
  qed
qed
end

fun shows_monom :: "('v :: show)monom \<Rightarrow> string \<Rightarrow> string"
where "shows_monom [(x,p)] = (if p = 1 then shows x else shows x +@+ shows_string ''^'' +@+ shows p)"
   |  "shows_monom ((x,p) # m) = ((if p = 1 then shows x else shows x +@+ shows_string ''^'' +@+ shows p) +@+ shows_string ''*'' +@+ shows_monom m)"
   |  "shows_monom [] = shows_string ''1''"

fun shows_poly :: "('v :: show,'a :: {poly_carrier,show})poly \<Rightarrow> string \<Rightarrow> string"
where "shows_poly [] = shows_string ''0''"
  | "shows_poly ((m,c) # p) = ((if c = 1 then shows_monom m else if m = [] then shows c else shows c +@+ shows_string ''*'' +@+ shows_monom m) +@+ 
           (if p = [] then shows_string [] else shows_string '' + '' +@+ shows_poly p))"

definition create_nlpoly_redpair :: "'a :: {show,poly_carrier} \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> bool \<Rightarrow> bool \<Rightarrow> ('f,'a)poly_inter_list \<Rightarrow> ('f :: show,'v :: show)redpair_impl"
where "create_nlpoly_redpair def gt power_mono discrete I = (let 
   J = poly_inter_list_to_inter def I;
   x = poly_subst (\<lambda> n. poly_of (PVar (''x_'' @ (shows n []))))
   in 
    (check_poly_inter_list discrete I <+? (\<lambda> (f,p). shows_string ''interpretation '' +@+ shows_poly (x p) +@+ shows_string '' of '' +@+ shows f +@+ shows_string '' invalid ''),
     (\<lambda> (s,t). let p = eval_term J s; q = eval_term J t in check (check_poly_gt gt p q) (shows_string ''could not ensure '' +@+ shows_poly p +@+ shows_string '' > '' +@+ shows_poly q)),
     (\<lambda> (s,t). let p = eval_term J s; q = eval_term J t in check (check_poly_ge p q) (shows_string ''could not ensure '' +@+ shows_poly p +@+ shows_string '' >= '' +@+ shows_poly q)),
     poly_inter_to_af J,
     check_all (\<lambda> ((f,n),p). list_all (\<lambda> i. check_poly_strict_mono_smart discrete power_mono gt p i) [0 ..< n]) I <+? (\<lambda> ((f,n),p). shows_string ''could not ensure monotonicty of '' +@+ shows_poly (x p) +@+ 
 shows_string '' as interpretation of '' +@+ shows f),
     shows_string ''polynomial interpretation'' +@+ shows_nl +@+ shows_sep (\<lambda> ((f,n),p) . 
        shows_string ''Pol('' +@+ shows f +@+ shows_string ''/'' +@+ shows n +@+ shows_string '') = '' +@+ shows_poly (x p)) shows_nl I ))"

lemma poly_order_carrier_to_generic_redpair_impl: 
  fixes default :: "'a :: {show,poly_carrier}"
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and power_mono :: bool
  and discrete :: bool
  assumes "poly_order_carrier default gt power_mono discrete"
  shows "generic_redpair_impl (create_nlpoly_redpair default gt power_mono discrete)"
proof -
  from assms interpret poly_order_carrier default gt power_mono discrete .
  show ?thesis
  proof
    fix I :: "('f :: show,'a)poly_inter_list" and s_list ns_list :: "('f,'v :: show)rule list" and type :: "'v"
    let ?rp = "create_nlpoly_redpair default gt power_mono discrete I :: ('f,'v)redpair_impl"
    let ?af = "rp_af ?rp :: 'f af"
    assume valid: "isOK (rp_valid ?rp)" 
    assume s: "isOK (check_allm (rp_strict ?rp) s_list)" 
    assume ns: "isOK (check_allm (rp_non_strict ?rp) ns_list)"
    from valid have valid: "isOK(check_poly_inter_list discrete I)" unfolding create_nlpoly_redpair_def by (auto simp: Let_def)
    let ?J = "poly_inter_list_to_inter default I :: ('f,'a)poly_inter"
    from ce_poly_order[OF valid] interpret ce_poly_order default gt power_mono discrete ?J type ?af unfolding create_nlpoly_redpair_def by (simp add: Let_def) 
    have mrp: "mono_ce_af_redpair (inter_s :: ('f,'v)trs) inter_ns ?af" ..
    have rp: "ce_af_redpair (inter_s :: ('f,'v)trs) inter_ns ?af" ..
    show "\<exists> S NS. ce_af_redpair S NS (rp_af ?rp) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK(rp_mono ?rp)  \<longrightarrow> mono_ce_af_redpair S NS (rp_af ?rp) \<and> one_step_ctxt_closed S)"
    proof (intro exI allI conjI, rule rp)
      show "set s_list \<subseteq> inter_s" 
      proof
        fix s t
        assume "(s,t) \<in> set s_list"
        with s have "check_poly_gt gt (eval_term ?J s) (eval_term ?J t)" unfolding create_nlpoly_redpair_def
          by (auto simp: Let_def)
        from check_poly_gt[OF this] show "(s,t) \<in> inter_s" unfolding inter_s_def by simp
      qed
    next
      show "set ns_list \<subseteq> inter_ns" 
      proof
        fix s t
        assume "(s,t) \<in> set ns_list"
        with ns have "check_poly_ge (eval_term ?J s) (eval_term ?J t)" unfolding create_nlpoly_redpair_def
          by (auto simp: Let_def)
        from check_poly_ge[OF this] show "(s,t) \<in> inter_ns" unfolding inter_ns_def by simp
      qed
    next
      show " isOK (rp_mono ?rp) \<longrightarrow>
        mono_ce_af_redpair inter_s inter_ns ?af \<and>
        one_step_ctxt_closed inter_s"
      proof (intro impI conjI, simp only: mrp)
        assume mono: "isOK(rp_mono ?rp)"
        interpret mono_poly_order default gt power_mono discrete ?J type ?af
        proof
          fix f :: 'f and  n i :: nat
          assume i: "i < n" 
          show "poly_strict_mono type (?J f n) i"
          proof (cases "lookup (f,n) I")
            case None
            hence d: "?J f n = default_I default n" unfolding poly_inter_list_to_inter_def by auto
            show ?thesis 
              by (simp only: d, rule check_poly_strict_mono, rule default_I_check_poly_strict_mono[OF i], rule default_I_check_poly_weak_mono[OF default_ge_zero])
          next
            case (Some p)
            hence p: "?J f n = p" and fnp: "((f,n),p) \<in> set I" unfolding poly_inter_list_to_inter_def using lookup_sound[OF Some] by auto
            show ?thesis
            proof (simp only: p, 
                rule check_poly_strict_mono_smart[OF mono[unfolded create_nlpoly_redpair_def, simplified Let_def, simplified, unfolded check_poly_inter_list_def list_all_iff, THEN bspec[OF _ fnp], simplified, THEN bspec]])
              show "i \<in> {0..<n}" by (simp add: i)
            next 
              show "check_poly_weak_mono_and_pos discrete p"
                by (rule valid[unfolded create_nlpoly_redpair_def, simplified Let_def, simplified, unfolded check_poly_inter_list_def, simplified,  THEN bspec[OF _ fnp], simplified])
            qed
          qed
        qed
        show "one_step_ctxt_closed inter_s" by (rule inter_s_mono)
      qed
    qed
  qed
qed

(* the following line did not work due to additional type constraint that 'a must be of sort show *)
(* sublocale poly_order_carrier \<subseteq> generic_redpair_impl "create_nlpoly_redpair default gt power_mono" *)


end
  
